import os
from . import sections
from .. project import aliases, importer, merge
from .. util import class_name, data_file, log


class Description:
    """
    Set and get project description, either as attributes or as indexes.

    Examples are :

    desc.run['threaded'] = True
    desc.animation = '.tests.PixelTester'
    """

    def __init__(self):
        self.clear()

    def clear(self):
        """Clear description to default values"""
        self._desc = {}
        for key, value in merge.DEFAULT_PROJECT.items():
            if key not in self._HIDDEN:
                self._desc[key] = type(value)()

    def items(self):
        """Return an iterable of (key, section value) pairs"""
        return self._desc.items()

    def update(self, desc=None, **kwds):
        """This method updates the description much like dict.update(), *except*:

        1. for description which have dictionary values, it uses update
           to alter the existing value and does not replace them.

        2. `None` is a special value that means "clear section to default" or
           "delete field".
        """
        sections.update(self._desc, desc, **kwds)

    def as_dict(self):
        """Returns a dictionary of non-empty description"""
        return {k: v for k, v in self.items() if v}

    @property
    def desc(self):
        return self._desc

    def __str__(self):
        return data_file.dumps(self.as_dict()).strip()

    _HIDDEN = 'maker', 'typename'
    _ATTRIBUTES = ()

    def __getitem__(self, key):
        return self._desc[key]

    def __setitem__(self, key, value):
        sections.set_one(self._desc, key, value)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError("'Description' has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        if key.startswith('_') or key in self._ATTRIBUTES:
            super().__setattr__(key, value)
        else:
            self[key] = value

    def __dir__(self):
        return sorted(super().__dir__() + list(self._desc))

    def __contains__(self, key):
        return key in self._desc

    def __iter__(self):
        return iter(self._desc)



import traceback, weakref
from .. util import log
from .. util.threads import runnable


class Runner:
    """
    Class that makes sure that a single builder is running.
    """
    def __init__(self, builder):
        self.thread = None
        self.is_running = False
        self.builder = weakref.ref(builder)

    def start(self, threaded):
        """Creates and starts the project."""
        self.stop()
        self.__class__._INSTANCE = weakref.ref(self)
        self.is_running = True

        if threaded:
            self.thread = runnable.LoopThread()
            self.thread.run_once = self._target
            self.thread.start()
        else:
            self._target()

    def stop(self):
        """
        Stop the Runner if it's running.
        Called as a classmethod, stop the running instance if any.
        """
        if self.is_running:
            log.info('Stopping')
            self.is_running = False
            self.__class__._INSTANCE = None

            try:
                self.thread and self.thread.stop()
            except:
                log.error('Error stopping thread')
                traceback.print_exc()
            self.thread = None
            return True

    @classmethod
    def instance(cls):
        """Return the unique instance of Runner, if any, or None"""
        return cls._INSTANCE and cls._INSTANCE()

    _INSTANCE = None

    def _target(self):
        try:
            self.builder()._run()
        except:
            if self.thread:
                log.error(traceback.format_exc())
            raise
        finally:
            b = self.builder()
            b and b.stop()



import contextlib, traceback


@contextlib.contextmanager
def add(*args):
    """
    A context manager that appends arguments to any exception thrown

    :param args: Arguments to be appended to the ``.args`` attribute of any
                 exception that is thrown while the context manager is active
    """
    try:
        yield
    except Exception as e:
        e.args = args + e.args
        raise


def report(function, *args, **kwds):
    """Run a function, catch, report and discard exceptions"""
    try:
        function(*args, **kwds)
    except Exception:
        traceback.print_exc()



import operator, os, re
from .. util import log

ALIAS_MARKERS = '@$'
SEPARATORS = re.compile(r'([./#]|[^./#]+)')

PROJECT_ALIASES = {}

BUILTIN_ALIASES = {
    'apa102': 'bibliopixel.drivers.SPI.APA102.APA102',
    'lpd8806': 'bibliopixel.drivers.SPI.LPD8806.LPD8806',
    'pi_ws281x': 'bibliopixel.drivers.PiWS281X.PiWS281X',
    'serial': 'bibliopixel.drivers.serial.Serial',
    'sk9822': 'bibliopixel.drivers.SPI.APA102.APA102',
    'spi': 'bibliopixel.drivers.SPI.SPI',
    'ws2801': 'bibliopixel.drivers.SPI.WS2801.WS2801',
    'ws281x': 'bibliopixel.drivers.SPI.WS281X.WS281X',

    'bpa': 'BiblioPixelAnimations',
}


def get_alias(alias):
    return PROJECT_ALIASES.get(alias) or BUILTIN_ALIASES.get(alias)


def resolve(typename, aliases=None):
    aliases = aliases or {}

    def get(s):
        return aliases.get(s) or get_alias(s) or s

    def get_all(typename):
        for part in SEPARATORS.split(typename):
            is_alias = part and (part[0] in ALIAS_MARKERS)
            yield get(part[1:]) if is_alias else part

    return ''.join(get_all(get(typename)))



"""
Run specified project from file or URL
"""

import os, string, sys, time, traceback
from .. main import project_flags
from .. util import data_file, log
from .. animation.animation import Animation
from .. project import load
from .. project.aliases import ALIAS_MARKERS
from .. project.project import Project

MOVIE_SUFFIXES = '.gif', '.mp4', '.yml'
_RUNNING = False


def stop():
    global _RUNNING
    _RUNNING = False
    Project.stop_all()


def run(args):
    Animation.FAIL_ON_EXCEPTION = args.fail_on_exception

    args.name = args.name or ['']
    projects = _get_projects(args)

    if args.dry_run:
        log.printer('(dry run - nothing executed)')
        return

    global _RUNNING
    _RUNNING = True

    needs_pause = False
    assert len(projects) == len(args.name)
    is_movie = (args.movie != '')

    for project, name in zip(projects, args.name):
        if not _RUNNING:
            break

        if needs_pause:
            if args.pause:
                time.sleep(float(args.pause))
                if not _RUNNING:
                    break
        else:
            needs_pause = True

        log.debug('Running file %s', name)
        if is_movie:
            project.flat_out()

        try:
            project.run()

        except KeyboardInterrupt:
            if not args.ignore_exceptions:
                raise
            log.warning('\nKeyboardInterrupt terminated project.')
            needs_pause = False

        except Exception as e:
            if not args.ignore_exceptions:
                raise
            log.error('Exception %s', e)
            traceback.print_exc()


def _load_py(filename):
    module = load.module(filename)
    project = module.get('PROJECT', {})

    if 'animation' not in project:
        no_extension = os.path.splitext(filename)[0]
        module_name = os.path.split(no_extension)[-1]
        names = module.keys()
        try:
            animation_name = load.guess_name(names, module_name, filename)
        except:
            if 'Animation' not in names:
                raise ValueError('Cannot deduce animation in file ' + filename)
            animation_name = 'Animation'
        animation = module[animation_name]
        if not callable(animation):
            raise ValueError('Animation "%s" in file "%s" is not callable'
                             % (animation_name, filename))

        project['animation'] = {'datatype': animation}

    return project


def _dump(args, desc):
    return data_file.dumps(desc, use_yaml=not args.json, safe=False)


def _get_projects(args):
    projects, failed = [], []

    movie = args.gif if args.movie == '' else args.movie
    if movie != '':
        log.info('writing animated movies')
        if movie is None:
            # --movie appeared but without a value
            movie = {}
        elif '{' in movie:
            movie = data_file.loads(movie, filename='--movie')

        if isinstance(movie, str):
            movie = {'filename': movie}
        filename = movie.get('filename', None)
        if filename:
            if not any(filename.endswith(s) for s in MOVIE_SUFFIXES):
                log.warning('--movie argument %s didn\'t end with %s',
                            movie, ' or '.join(MOVIE_SUFFIXES))
                movie += '.gif'
            if len(args.name) > 1:
                log.warning('Writing multiple movies to the same file name')

        movie['typename'] = '.movie_writer'

    # Handle spaces around + signs.
    names = []
    for name in args.name:
        if names and (name.startswith('+') or names[-1].endswith('+')):
            names[-1] += name
        else:
            names.append(name)
    args.name = names

    for project_files in args.name:
        saved_path = sys.path[:]
        descs = []
        root_file = None
        try:
            for filename in project_files.split('+'):
                if filename.endswith('.py'):
                    desc = _load_py(filename)
                    root_file = filename
                elif '{' in filename:
                    desc = data_file.loads(filename)
                elif '.' in filename:
                    desc = load.data(filename, False)
                    desc = data_file.load(desc, filename)
                    root_file = filename
                else:
                    raise ValueError('Do not understand command ' + filename)

                descs.append(desc)

            if movie:
                for d in descs:
                    d.pop('driver', None)
                if 'filename' in movie:
                    driver = movie
                else:
                    f = project_files.replace('+', '-')
                    f = f.replace('/', '-')
                    f = ''.join(i for i in f if i in _FILE_CHARS)
                    f, ext = os.path.splitext(f)
                    driver = dict(movie, filename=f + '.gif')
                descs.append({'driver': driver})

            project = project_flags.make_project(
                args, *descs, root_file=root_file)
            projects.append(project)
            if args.dump:
                log.printer(_dump(args, project.desc))

        except FileNotFoundError as e:
            failed.append(('%s: %s' % (e.strerror, e.filename), ()))

        except Exception as exception:
            if filename.endswith('.py'):
                raise
            eargs = exception.args
            desc = _dump(args, descs)
            if args.verbose:
                exception = traceback.format_exc()
            msg = RUN_ERROR.format(**locals())
            failed.append((msg, eargs))

        finally:
            sys.path[:] = saved_path

    if not failed:
        return projects

    log.error(FAILURE_ERROR.format(
        count=len(failed), s='' if len(failed) == 1 else 's'))
    for msg, args in failed:
        if args:
            log.error('\n'.join(str(a) for a in args) + '\n')
        else:
            log.error(msg + '\n')
    raise ValueError('Run aborted')


_ALL_CHARS = set(string.ascii_letters + string.digits + '._' + ALIAS_MARKERS)
_START_CHARS = set(string.ascii_letters + '.' + ALIAS_MARKERS)
_END_CHARS = set(string.ascii_letters + string.digits)
_FILE_CHARS = set(string.ascii_letters + string.digits + '._-')


RUN_ERROR = """When reading file {filename}:

{desc}
{exception}
"""

FAILURE_ERROR = """\
{count} project{s} failed
____________________
"""


def _might_be_classname(s):
    return (s and
            set(s) <= _ALL_CHARS and
            s[0] in _START_CHARS and
            s[-1] in _END_CHARS)



import os, pathlib

DIR = os.listdir(str(pathlib.Path(__file__).parent))
COMMANDS = [i[:-3] for i in DIR if i.endswith('.py') and not i.startswith('_')]
COMMANDS_LINES = COMMANDS[0:8], COMMANDS[8:]
COMMANDS_PRINTABLE = '\n'.join(', '.join(i) for i in COMMANDS_LINES)



import os
from .. project import aliases, importer
from .. util import class_name, data_file, log

CLASS_SECTIONS = 'layout', 'animation', 'driver'


def set_one(desc, name, value):
    """Set one section in a Project description"""
    old_value = desc.get(name)
    if old_value is None:
        raise KeyError('No section "%s"' % name)

    if value is None:
        value = type(old_value)()

    elif name in CLASS_SECTIONS:
        if isinstance(value, str):
            value = {'typename': aliases.resolve(value)}
        elif isinstance(value, type):
            value = {'typename': class_name.class_name(value)}
        elif not isinstance(value, dict):
            raise TypeError('Expected dict, str or type, got "%s"' % value)

        typename = value.get('typename')
        if typename:
            s = 's' if name == 'driver' else ''
            path = 'bibliopixel.' + name + s
            importer.import_symbol(typename, path)

    elif name == 'shape':
        if not isinstance(value, (list, int, tuple, str)):
            raise TypeError('Expected shape, got "%s"' % value)

    elif type(old_value) is not type(value):
        raise TypeError('Expected %s but got "%s" of type %s' %
                        (type(old_value), value, type(value)))

    desc[name] = value


def update(desc, other=None, **kwds):
    """Update sections in a Project description"""
    other = other and _as_dict(other) or {}
    for i in other, kwds:
        for k, v in i.items():
            if isinstance(v, dict):
                # Only for dicts, merge instead of overwriting
                old_v = desc[k]
                for k2, v2 in v.items():
                    if v2 is None:
                        old_v.pop(k2, None)
                    else:
                        old_v[k2] = v2
            else:
                set_one(desc, k, v)


def _as_dict(data):
    if isinstance(data, dict):
        return data

    as_dict = getattr(data, 'as_dict', None)
    if callable(as_dict):
        return as_dict()

    if isinstance(as_dict, dict):
        return as_dict

    desc = getattr(data, 'desc', None)
    if desc is not None:
        return _as_dict(desc)

    if isinstance(data, str):
        data = data_file.load_if(data)
        if isinstance(data, dict):
            return data

    raise TypeError('Expected dict but got value "%s"' % data)



import os
from .. project import aliases, construct, load, project
from . import animation, failed, runner
from .. util import log

DEFAULT_ANIMATION = '.animation'
ANIMATION_PATH = 'bibliopixel.animation'


class Collection(animation.Animation):
    """
    A ``Collection`` is a list of ``Animation``s
    """

    @staticmethod
    def pre_recursion(desc):
        animations = desc['animations']
        animations = [_clean_animation(a, desc) for a in animations]
        _make_names_unique(animations)
        desc['animations'] = animations
        return desc

    CHILDREN = 'animations',

    def __init__(
            self, layout, animations=None, **kwds):
        super().__init__(layout, **kwds)
        self.animations = _AnimationList(animations or [])
        self.internal_delay = 0  # never wait
        for a in self.animations:
            a.top_level = False

    def cleanup(self, clean_layout=True):
        self.state = runner.STATE.canceled
        for a in self.animations:
            a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        from .. util import deprecated
        deprecated.deprecated('Collection.add_animation')

        anim._set_runner(kwds)
        self.animations.append(anim)

    def pre_run(self):
        for a in self.animations:
            a.pre_run()

    def set_project(self, project):
        super().set_project(project)
        for a in self.animations:
            a.set_project(project)

    def detach(self, overlay):
        """
        Give each animation a unique, mutable layout so they can run
        independently.
        """
        # See #868
        for i, a in enumerate(self.animations):
            a.layout = a.layout.clone()
            if overlay and i:
                a.preclear = False


class _AnimationList:
    def __init__(self, animations):
        self._animations = list(animations)
        self._names = {a.name: i for i, a in enumerate(self._animations)}

    def _index(self, i):
        if isinstance(i, int):
            return i % len(self._animations)
        return self._names[i]

    def append(self, animation):
        # We're forced to do the "fill in a name and make sure it's
        # unique" in two different ways, because of this method.
        #
        # TODO: We should disallow APIs from doing surgery on
        # self.animations after construction.  Right now only
        # Remote does this, or legacy code that calls add_animation.
        # That code should go on ``cls.pre_recursion()``.
        self._animations.append(animation)
        try:
            base_name = animation.name
        except AttributeError:
            base_name = animation.__class__.__name__

        count = 0
        name = base_name

        while name in self._names:
            name = '%s_%d' % (base_name, count)
            count += 1

        animation.name = name

    def __getitem__(self, i):
        return self._animations[self._index(i)]

    def __getattr__(self, i):
        return self[i]

    def __iter__(self):
        return iter(self._animations)

    def __len__(self):
        return len(self._animations)


def _clean_animation(desc, parent):
    """
    Cleans up all sorts of special cases that humans want when entering
    an animation from a yaml file.

    1. Loading it from a file
    2. Using just a typename instead of a dict
    3. A single dict representing an animation, with a run: section.
    4. (Legacy) Having a dict with parallel elements run: and animation:
    5. (Legacy) A tuple or list: (animation, run )

    """
    desc = load.load_if_filename(desc) or desc

    if isinstance(desc, str):
        animation = {'typename': desc}

    elif not isinstance(desc, dict):
        raise TypeError('Unexpected type %s in collection' % type(desc))

    elif 'typename' in desc or 'animation' not in desc:
        animation = desc

    else:
        animation = desc.pop('animation', {})
        if isinstance(animation, str):
            animation = {'typename': animation}

        animation['run'] = desc.pop('run', {})
        if desc:
            raise ValueError('Extra animation fields: ' + ', '.join(desc))

    animation.setdefault('typename', DEFAULT_ANIMATION)
    animation = construct.to_type_constructor(animation, ANIMATION_PATH)
    datatype = animation.setdefault('datatype', failed.Failed)
    animation.setdefault('name', datatype.__name__)

    # Children without fps or sleep_time get it from their parents.
    # TODO: We shouldn't have to rewrite our descriptions here!  The
    # animation engine should be smart enough to figure out the right
    # speed to run a subanimation without a run: section.
    run = animation.setdefault('run', {})
    run_parent = parent.setdefault('run', {})
    if not ('fps' in run or 'sleep_time' in run):
        if 'fps' in run_parent:
            run.update(fps=run_parent['fps'])
        elif 'sleep_time' in run_parent:
            run.update(sleep_time=run_parent['sleep_time'])

    return animation


def _make_names_unique(animations):
    """
    Given a list of animations, some of which might have duplicate names, rename
    the first one to be <duplicate>_0, the second <duplicate>_1,
    <duplicate>_2, etc."""
    counts = {}
    for a in animations:
        c = counts.get(a['name'], 0) + 1
        counts[a['name']] = c
        if c > 1:
            a['name'] += '_' + str(c - 1)

    dupes = set(k for k, v in counts.items() if v > 1)
    for a in animations:
        if a['name'] in dupes:
            a['name'] += '_0'



from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    from . animation import Animation, BaseAnimation
    from . circle import BaseCircleAnim, Circle
    from . cube import BaseCubeAnim
    from . game import BaseGameAnim
    from . matrix import BaseMatrixAnim
    from . off import Off, OffAnim
    from . receiver import BaseReceiver
    from . sequence import Sequence
    from . strip import BaseStripAnim
    from . tests import (
        StripChannelTest, MatrixChannelTest, MatrixCalibrationTest)

    strip_test = StripChannelTest
    matrix_calibration = MatrixCalibrationTest
    matrix_test = MatrixChannelTest



from . pixels import Pixels
from . runner import Runner
from . saved_description import SavedDescription
from .. import colors
from .. commands import animations
from .. drivers.SimPixel import driver as simpixel_driver
from .. project import project
from .. util import log


class Builder(SavedDescription):
    """
    A Project Builder to allow people to experiment with projects
    from the command line or in their own main program.

    """
    COLORS = colors.COLORS

    def __init__(self, project_file='', desc=None, threaded=None, **kwds):
        self.project = None
        self.pixel = Pixels(self)
        self._runner = Runner(self)

        super().__init__(project_file, desc, **kwds)

        if threaded is not None:
            self.threaded = threaded

    def start(self, threaded=None):
        """Creates and starts the project."""
        if threaded is not None:
            self.threaded = threaded
        run = {'run': {'threaded': False}}
        self.project = project.project(
            self.desc, run, root_file=self.project_file)
        self._run = self.project.run
        self._runner.start(self.threaded)

    def stop(self=None):
        """Stop the builder if it's running."""
        if not self:
            instance = getattr(Runner.instance(), 'builder', None)
            self = instance and instance()
            if not self:
                return

        self._runner.stop()
        if self.project:
            self.project.stop()
            self.project = None

    def clear(self):
        """Stop the project if it's running and clear the project description"""
        self.stop()
        super().clear()

    @property
    def is_running(self):
        """True if the Builder is currently running"""
        return self._runner.is_running

    @property
    def threaded(self):
        """
        True if the Builder is runs in a separate thread, false if the
        Builder blocks, waiting for the animation to end.
        """
        return self.desc.run.get('threaded', False)

    @threaded.setter
    def threaded(self, value):
        self.desc.run['threaded'] = bool(value)

    @staticmethod
    def simpixel(new=0, autoraise=True):
        """Open an instance of simpixel in the browser"""
        simpixel_driver.open_browser(new=new, autoraise=autoraise)

    @staticmethod
    def animations():
        """List all the existing animations"""
        animations.run(None)

    def __repr__(self):
        r = super().__repr__()
        return 'Running ' + r if self.is_running else r

    def __iadd__(self, other):
        self.desc.update(other)
        return self

    def __add__(self, other):
        b = Builder()
        b += self
        b += other
        return b

    def __radd__(self, other):
        b = Builder()
        b += other
        b += self
        return b

    def __copy__(self):
        return self + {}

    def __deepcopy__(self, _):
        return self.__copy__()
